from unittest import TestCase

from solax_py_library.device.types.device import DeviceType
from solax_py_library.smart_scene.core.condition import (
    DateCondition,
    BaseCondition,
    CabinetCondition,
)
from solax_py_library.smart_scene.types.condition import (
    CabinetConditionItemData,
    CabinetConditionType,
    ConditionItemChildData,
)
from solax_py_library.smart_scene.types.condition_value import CabinetValue


class TestCondition(TestCase):
    def test_condition_build(self):
        date_condition = DateCondition(
            update_value_function=lambda: 1,
        )
        assert isinstance(date_condition, BaseCondition)

    def test_cabinet_condition_to_text(self):
        cabinet_condition = CabinetConditionItemData(
            childType=CabinetConditionType.cabinetAlarm,
            childData=ConditionItemChildData(
                data=[DeviceType.IO_TYPE, DeviceType.COLD_TYPE, 1]
            ),
        )
        print(cabinet_condition.to_text(lang="zh_CN", unit="嘻嘻"))

    def test_cabinet_condition_check(self):
        cabinet_condition = CabinetCondition(
            update_value_function=lambda: {
                "SN1": CabinetValue(
                    soc=0,
                    io_alarm=[False, True, False],
                ),
            },
        )
        cabinet_condition.update_value()
        assert (
            cabinet_condition.meet_func(
                data=CabinetConditionItemData(
                    childType=CabinetConditionType.cabinetAlarm,
                    childData=ConditionItemChildData(
                        data=[DeviceType.IO_TYPE, DeviceType.COLD_TYPE, 1]
                    ),
                ),
                ctx={"cabinet": ["SN1"]},
            )
            is False
        )
        assert (
            cabinet_condition.meet_func(
                data=CabinetConditionItemData(
                    childType=CabinetConditionType.cabinetAlarm,
                    childData=ConditionItemChildData(
                        data=[DeviceType.IO_TYPE, DeviceType.COLD_TYPE, 2]
                    ),
                ),
                ctx={"cabinet": ["SN1"]},
            )
            is True
        )



from typing import List, Dict, Any

from solax_py_library.smart_scene.types.action import SmartSceneAction, ActionType
from solax_py_library.smart_scene.types.condition import (
    SmartSceneCondition,
    ConditionType,
)


def action_param_check(actions: List[SmartSceneAction], ctx: Dict[str, Any]):
    """动作里的参数范围判定"""
    for action in actions:
        if action.type != ActionType.system:
            continue
        for action_data in action.data:
            ret = action_data.check_param(ctx)
            if ret is not None:
                return ret
    return True


def condition_param_check(condition: SmartSceneCondition, ctx: Dict[str, Any]):
    for condition_data in condition.value:
        if condition_data.type not in [
            ConditionType.systemCondition,
            ConditionType.cabinet,
        ]:
            continue
        for data in condition_data.data:
            ret = data.check_param(ctx)
            if ret is not None:
                return ret
    return True



class BaseCondition(object):
    condition_type = None

    def __init__(self, update_value_function, **kwargs):
        self.value = {}
        if not callable(update_value_function):
            raise ValueError("update_value_function must be callable")
        self.update_value_function = update_value_function

    def update_value(self):
        new_value = self.update_value_function()
        if not isinstance(new_value, dict):
            raise ValueError("update_value_function must return a dict")
        self.value.update(new_value)

    def meet_func(self, data, ctx):
        ...



from datetime import datetime

from tornado.log import app_log

from solax_py_library.smart_scene.core.condition.base import BaseCondition
from solax_py_library.smart_scene.types.condition import (
    PriceConditionItemData,
    PriceConditionType,
    SmartSceneUnit,
    ConditionFunc,
    ConditionType,
)
from solax_py_library.utils.time_util import (
    get_left_timestamp,
    hour_to_today_timestamp,
)


class ElePriceCondition(BaseCondition):
    def __init__(self, update_value_function, **kwargs):
        super().__init__(update_value_function, **kwargs)
        self.buy = None

    def meet_func_price(
        self, function_value: ConditionFunc, data_value, current_data
    ) -> bool:
        """电价条件的判定"""
        app_log.info(f"meet_func_price: {current_data}, data_value: {data_value[0]}")
        return function_value.function()(current_data, data_value[0])

    def meet_func_highest_price(self, data_value, current_data) -> bool:
        value, unit = data_value
        price = self.value["price"]
        max_num = max(price.values())
        if unit == SmartSceneUnit.NUM:  # 比最高电价低X元
            base = round(max_num - value, 5)
        else:  # 比最高电价低X%
            if max_num < 0:
                base = round(max_num * (1 + value / 100), 5)
            else:
                base = round(max_num * (1 - value / 100), 5)
        app_log.info(f"meet_func_highest_price: {base}, data_value: {current_data}")
        if current_data <= base:
            return True
        else:
            return False

    def meet_func_lowest_price(self, data_value, current_data) -> bool:
        value, unit = data_value
        price = self.value["price"]
        min_num = min(price.values())
        if unit == SmartSceneUnit.NUM:  # 比最低电价高X元
            base = round(min_num + value, 5)
        else:  # 比最低电价高X%
            if min_num < 0:
                base = round(min_num * (1 - value / 100), 5)
            else:
                base = round(min_num * (1 + value / 100), 5)
        app_log.info(f"meet_func_lowest_price: {base}, data_value: {current_data}")
        if current_data >= base:
            return True
        else:
            return False

    def meet_func_highest_or_lowest_hours(
        self, data_value, current_timestamp, reverse
    ) -> bool:
        start_time = data_value[0]
        end_time = data_value[1]
        hours = data_value[2]
        price = self.value["price"]
        start_timestamp = hour_to_today_timestamp(int(start_time.split(":")[0]))
        end_timestamp = hour_to_today_timestamp(int(end_time.split(":")[0]))
        ret = []
        for timestamp_key, price_value in price.items():
            if (
                start_timestamp > int(timestamp_key)
                or int(timestamp_key) > end_timestamp
            ):
                continue
            ret.append([timestamp_key, price_value])
        sorted_price = sorted(ret, key=lambda x: x[1], reverse=reverse)
        top_sorted_price = sorted_price[: hours * 4]
        if current_timestamp in top_sorted_price:
            return True
        return False

    def meet_func(self, data: PriceConditionItemData, ctx):
        if not self.value or not self.value.get("price"):
            # 未获取到价格数据，直接返回
            return False
        child_data = data.childData
        child_type = data.childType
        data_value = child_data.data
        now_time = int(datetime.now().timestamp() * 1000)
        current_timestamp = get_left_timestamp(now_time)
        if str(current_timestamp) not in self.value["price"]:
            return False
        current_data = self.value["price"][str(current_timestamp)]
        if child_type == PriceConditionType.price:
            return self.meet_func_price(child_data.function, data_value, current_data)
        elif child_type == PriceConditionType.lowerPrice:
            return self.meet_func_highest_price(data_value, current_data)
        elif child_type == PriceConditionType.higherPrice:
            return self.meet_func_lowest_price(data_value, current_data)
        elif child_type == PriceConditionType.expensiveHours:
            return self.meet_func_highest_or_lowest_hours(
                data_value, current_timestamp, reverse=True
            )
        elif child_type == PriceConditionType.cheapestHours:
            return self.meet_func_highest_or_lowest_hours(
                data_value, current_timestamp, reverse=False
            )
        return False


class EleSellPriceCondition(ElePriceCondition):
    condition_type = ConditionType.sellingPrice

    def __init__(self, update_value_function, **kwargs):
        super().__init__(update_value_function, **kwargs)
        self.buy = False


class ElsBuyPriceCondition(ElePriceCondition):
    condition_type = ConditionType.buyingPrice

    def __init__(self, update_value_function, **kwargs):
        super().__init__(update_value_function, **kwargs)
        self.buy = True



from .base import BaseUploadService
from .ftp import FTPUploadService


upload_service_map = {}


def _register(upload_obj):
    upload_service_map[upload_obj.upload_type] = upload_obj


_register(FTPUploadService)


__all__ = ["BaseUploadService", "FTPUploadService"]



import ftplib
import os

from .base import BaseUploadService
from solax_py_library.upload.types.client import UploadType
from solax_py_library.upload.core.data_adapter import CSVDataAdapter
from solax_py_library.upload.types.ftp import (
    FTPServiceConfig,
    FTPFileType,
    FTPData,
    FTPParsedData,
)
from solax_py_library.upload.exceptions.upload_error import (
    SendDataError,
    ConnectError,
    LoginError,
    ConfigurationError,
)


class FTPUploadService(BaseUploadService):
    upload_type = UploadType.FTP

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        config = self._init_config(**kwargs)
        self.host = config.host
        self.port = config.port
        self.user = config.user
        self.password = config.password
        self.remote_path = config.remote_path

    def _init_config(self, **kwargs):
        try:
            return FTPServiceConfig(**kwargs)
        except Exception as e:
            raise ConfigurationError(message=str(e))

    def connect(self):
        if self._client is not None:
            self.close()
        self._connect()
        self._login()
        self._is_connect = True

    def _connect(self):
        try:
            self._client = ftplib.FTP()
            self._client.encoding = "utf8"  # 默认需要配置为utf8
            self._client.connect(self.host, self.port, timeout=self.timeout)
        except Exception as e:
            raise ConnectError(message=str(e))

    def _login(self):
        try:
            self._client.login(self.user, self.password)
        except Exception as e:
            raise LoginError(message=str(e))

    def close(self):
        if self._client:
            self._client.quit()
            self._client = None
            self._is_connect = False
            print("Connection closed.")

    def _parse(self, upload_data: FTPData) -> FTPParsedData:
        if upload_data.file_type == FTPFileType.CSV:
            parsed_data = CSVDataAdapter.parse_data(upload_data.data)
        else:
            raise NotImplementedError
        return FTPParsedData(
            file_name=upload_data.file_name
            + FTPFileType.get_file_suffix(upload_data.file_type),
            file_path=parsed_data,
        )

    def _upload(self, data: FTPParsedData):
        try:
            if not self._is_connect:
                raise ConnectError(message="not connect yet")
            # self._client.set_pasv(False)
            self._cwd_folder()
            with open(data.file_path, "rb") as f:
                self._client.storbinary(f"STOR {data.file_name}", f)
            print(f"Successfully uploaded to {data.file_name}")
        except Exception as e:
            raise SendDataError(message=str(e))
        finally:
            if os.path.exists(data.file_path):
                os.remove(data.file_path)

    def _cwd_folder(self):
        try:
            self._client.cwd(self.remote_path)
        except Exception as e:
            print(f"CWD error: {e}")
            self._client.mkd(self.remote_path)
            self._client.cwd(self.remote_path)
            print(f"mkdir {self.remote_path}, remove successfully")

    def _build_remote_file(self, file_name):
        return os.path.join(self.remote_path, file_name)
